"""Take URL and return URL of archive.org snapshot.

Given a URL and (optionally) a timestamp, return the URL of the archive.org
snapshot closest to the provided timestamp. If no timestamp is provided or no
snapshot for the provided timestamp cannot be found, return the latest
snapshot. If no snapshot available, return None.
"""

import asyncio
import sys
from typing import Any, Optional

import httpx


async def call_api(client: httpx.AsyncClient, url: str,
                   timestamp: Optional[str] = None) -> dict[str, Any]:
    """Call Wayback Machine API and return JSON response.

    If API is unresponsive, sleep task for five seconds for a maximum of five times.
    After five unsuccessful attempts, exit program.

    Expect the following API responses:

    - URL with correctly formatted timestamp (YYYYMMDDhhmmss):
        JSON with nearest snapshot, latest snapshot if timestamp was not found
    - URL without timestamp:
        JSON with latest snapshot available
    - Any URL with badly formatted timestamp (e.g. 'May2000'):
        Empty JSON
    - URL that is has not available in archive.org:
        Empty JSON

    Args:
        client (httpx.AsyncClient): HTTPX AsyncClient to make API calls
        url (str): URL to be searched in the Wayback Machine
        timestamp (Optional[str]): Timestamp for desired snapshot

    Returns:
        dict[str, Any]: JSON API response
    """
    # Construct URL for API call
    api_call: str = f"https://archive.org/wayback/available?url={url}"
    if timestamp:
        api_call += f"&timestamp={timestamp}"
    # Call API
    tries: int = 0
    while tries < 5:
        try:
            response: httpx.Response = await client.get(api_call)
            response.raise_for_status()
            return response.json()
        except httpx.HTTPError:
            tries += 1
            print("Encountered HTTP error, waiting 5 seconds.")
            await asyncio.sleep(5)
            continue
    else:
        sys.exit("API unresponsive, try again later.")


def get_closest(api_response: dict[str, Any]) -> Optional[str]:
    """Get URL of closest snapshot from API response if available.

    Returns None if API response is empty.

    Args:
        api_response: dict[str, Any]: API response as JSON

    Returns:
        Optional[str]: URL of Wayback Machine snapshot, if any was found
    """
    if not api_response['archived_snapshots']:
        return None
    return api_response['archived_snapshots']['closest']['url']


async def gather_snapshots(
        urls: list[str], timestamp: str = None) -> dict[str, Optional[str]]:
    """Create HTTPX session for API calls and return gathered snapshots.

    To make asynchronous calls, create a task list for calling the call_api function
    and get results with asyncio.gather(). The completed tasks will then be used to
    build a dict of url-snapshot pairs.

    Args:
        urls (list[str]): Urls to send to the Wayback Machine API
        timestamp (Optional[str]): Timestamp to send to the Wayback Machine API

    Returns:
        dict[str, Optional[str]]: API call results with original URL as keys and Wayback
                                  snapshot URLs as values
    """
    async with httpx.AsyncClient(timeout=10) as client:
        # Create task list (with each task being an API call)
        tasks: list[asyncio.Task[Any]] = []
        for url in urls:
            tasks.append(asyncio.create_task(call_api(client, url, timestamp)))
        # Execute tasks and gather results
        api_responses: list[dict[str, Any]] = await asyncio.gather(*tasks)
        # Build url-snapshot pairs from results
        wayback_urls: dict[str, Optional[str]] = {}
        for api_response in api_responses:
            wayback_urls[api_response["url"]] = get_closest(api_response)
    return wayback_urls



"""Extract dates and URLs from Markdown files."""

from datetime import datetime
from pathlib import Path
from typing import Optional

import markdown
from bs4 import BeautifulSoup

# List of URLs considered stable and thus ignored
STABLE_URLS: tuple = (
    # archive.org snapshots
    "web.archive.org/web/",
    # Pelican intra-site links
    "{filename}",
    "{static}",
    # Jekyll intra-site links
    "{% post_url",
    # Hugo intra-site links
    "{{< ref",
    "{{< relref",
    # Persistent identifier list from ORCID (https://pub.orcid.org/v2.0/identifiers)
    "https://arxiv.org/abs/",
    "http://www.amazon.com/dp/",
    "https://www.authenticus.pt/",
    "http://adsabs.harvard.edu/abs/",
    "https://ciencia.iscte-iul.pt/id/",
    "https://d-nb.info/",
    "https://doi.org/",
    "http://ethos.bl.uk/OrderDetails.do?uin=",
    "https://hal.archives-ouvertes.fr/view/resolver/",
    "http://hdl.handle.net/",
    "https://www.worldcat.org/isbn/",
    "https://portal.issn.org/resource/ISSN/",
    "http://zbmath.org/?format=complete&q=an%3A",
    "http://www.jstor.org/stable/",
    "https://koreamed.org/article/",
    "http://lccn.loc.gov/",
    "https://www.lens.org/",
    "http://www.ams.org/mathscinet-getitem?mr=",
    "http://www.worldcat.org/oclc/",
    "http://openlibrary.org/b/",
    "https://www.osti.gov/biblio/",
    "http://identifiers.org/pdb/",
    "https://europepmc.org/articles/",
    "https://www.ncbi.nlm.nih.gov/pubmed/",
    "https://europepmc.org/article/PPR/",
    "https://tools.ietf.org/html/",
    "https://identifiers.org/rrid/",
    "http://papers.ssrn.com/abstract_id=",
    "http://zbmath.org/?format=complete&q="
)


def scan_md(md_source: str, md_file: Path) -> tuple[Optional[str], list[str]]:
    """Extract date and URLs from specified Markdown file.

    To get the date, first try to extract it from Markdown meta information. If no date
    found, try to extract date from file name by following the Jekyll naming convention
    where files for blog posts start with YYYY-MM-DD. Next, try to format date for
    Wayback Machine API as YYYYMMDD or YYYYMMDDhhmm.

    Args:
        md_source (str): Contents of the Markdown file
        md_file (Path): Markdown file path

    Returns:
        tuple[Optional[str], list[str]]: Formatted date (if found) and list of URLs
    """
    html, date = convert_markdown(md_source)
    if not date:
        date = md_file.name[:10]
    return format_date(date), filter_urls(get_urls(html))


def convert_markdown(md_source: str) -> tuple[str, Optional[str]]:
    """Convert Markdown file to HTML and extract date from metadata.

    Args:
        md_file (str): Contents of the Markdown file

    Returns:
        tuple[str, dict[str, Optional[str]]: HTML version of Markdown file and date from
                                             Markdown metadata
    """
    md: markdown.core.Markdown = markdown.Markdown(extensions=['meta'])
    html: str = md.convert(md_source)
    try:
        date: Optional[str] = md.Meta['date'][0]
    except KeyError:
        date = None
    return html, date


def format_date(date: str) -> Optional[str]:
    """Format date according to Wayback Machine API format.

    Recognizes dates in two formats: YYYY-MM-DD and YYYY-MM-DD hh:mm. These formats are
    returned as YYYYMMDD or YYYYMMDDhhmm. If format isn't recognized, return None.

    Args:
        date (str): Date extracted from Markdown metadata or file name

    Returns:
        Optional[str]: Date formatted as YYYYMMDDhhmm
    """
    try:
        # To capture dates with and without hour and minute, adjust strptime
        # format depending on length of split date string
        md_date: list[str] = date.split(" ")
        if len(md_date) == 2:
            # Date contains hour and minute
            formatted_date = datetime.strptime(date, "%Y-%m-%d %H:%M")
            return formatted_date.strftime("%Y%m%d%H%M")
        else:
            # Date only contains year, month and day
            formatted_date = datetime.strptime(date, "%Y-%m-%d")
            return formatted_date.strftime("%Y%m%d")
    # Malformatted date or no date at the beginning of file name
    except (KeyError, ValueError):
        return None


def filter_urls(md_urls: list[str]) -> list[str]:
    """Take and filter list of URLs for API calls.

    Filter out duplicates and remove URLs that are considered stable:

    - URLs that already point to archive.org
    - Intra-site links (recognizes Pelican, Jekyll and Hugo intra-site link formats)
    - URLs containing stable identifiers

    Args:
        md_urls (list[str]): List of URLs extracted from Markdown file

    Returns:
        list[str]: Filtered list of URLs
    """
    # Remove duplicates
    urls: list[str] = list(set(md_urls))
    # Filter out stable URLs
    return [url for url in urls if not
            any(stable_url in url for stable_url in STABLE_URLS)]


def get_urls(html: str) -> list[str]:
    """Extract links from converted Markdown HTML.

    Args:
        html (str): HTML version of Markdown file

    Returns:
        list[str]: URLs found in HTML
    """
    soup = BeautifulSoup(html, "html.parser")
    return [a.get('href') for a in soup.find_all('a', href=True)]



"""Command line interface."""

import argparse
import asyncio
import sys
from pathlib import Path

from archive_md_urls.update_files import update_files


def get_md_files(items: list[Path], recursive: bool) -> list[Path]:
    """Scan files and directories and create a list of Markdown files.

    For single file just check if file exists and has the correct file ending (.md).
    For directory, glob for files with .md ending and add each to file list.

    Args:
        item_list (list[str]): List of items provided via argparse
        recursive (bool): Recursively search subdirectories for Markdown files if True

    Returns:
        list[Path]: List of Markdown files to update
    """
    files: list[Path] = []
    for item in items:
        if item.is_dir():
            if recursive:
                files.extend(
                    list(item.rglob("**/*.md"))
                )
            else:
                files.extend(
                    list(item.glob("*.md"))
                )
        elif item.is_file():
            if item.suffix == ".md":
                files.append(item)
            else:
                sys.exit(f"No Markdown file extension (.md): {item}")
        else:
            sys.exit(f"Not a file or directory: {item}.")
    if not files:
        sys.exit("Couldn't find any Markdown files. Do you use the file ending .md for "
                 "your Markdown files? If yes, you could try to search directories "
                 "recursively using the -r flag (see help).")
    return files


def parse_args() -> argparse.Namespace:
    """Parse arguments."""
    argparser = argparse.ArgumentParser(
        description="Turn URLs in Markdown files into Wayback Machine snapshots."
    )
    argparser.add_argument(
        "items",
        nargs='+',
        type=Path,
        help="Markdown file, or directory containing Markdown files. You can specify "
             "multiple items and combine individual files with directories"
    )
    argparser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="Recursively search for Markdown files in subdirectories"
    )
    return argparser.parse_args()


def main() -> None:
    """archive-md-urls cli entry point."""
    args: argparse.Namespace = parse_args()
    files: list[Path] = get_md_files(args.items, args.recursive)
    asyncio.run(update_files(files))


if __name__ == "__main__":
    main()



"""Turn URLs in Markdown files to Wayback snapshots."""

from pathlib import Path
from typing import Optional

from archive_md_urls.gather_snapshots import gather_snapshots
from archive_md_urls.scan_md import scan_md


async def update_files(files: list[Path]) -> None:
    """Scan and update URLs in Markdown files.

    File contents are updated in-place.

    Args:
        files (list[Path]): List of Markdown files to scan and update
    """
    # Keep count of changed URLs to summarize changes to user
    changed_urls: int = 0
    for file in files:
        md_source: str = file.read_text(encoding="utf-8")
        # Call API and collect snapshots
        wayback_urls: dict[str, Optional[str]] = await get_snapshots(md_source, file)
        # Update links in file source and write file
        updated_md_source: str = update_md_source(md_source, wayback_urls)
        file.write_text(updated_md_source, encoding="utf-8")
        changed_urls += len([item for item in wayback_urls.values() if item])
    print(f"Changed {changed_urls} {'URL' if changed_urls == 1 else 'URLs'} "
          f"in {len(files)} {'file' if len(files) == 1 else 'files'}.")


async def get_snapshots(md_source: str, md_file: Path) -> dict[str, Optional[str]]:
    """Scan Markdown file and call Wayback Machine API to create list of url-snapshot pairs.

    Args:
        md_source (str): Contents of the Markdown file
        md_file (Path): Path of Markdown file

    Returns:
        dict[str, Optional[str]]: API call results with original URL as keys and Wayback
                                  snapshot URLs as values
    """
    date, urls = scan_md(md_source, md_file)
    return await gather_snapshots(urls, date)


def update_md_source(md_source: str, wayback_urls: dict[str, Optional[str]]) -> str:
    """Replace URLs in Markdown file with Wayback Snapshots.

    Args:
        md_source (str): Content of Markdown file that should be updated
        wayback_urls (dict[str, Optional[str]]): URL-Snapshot pairs

    Returns:
        str: Content of Markdown file with updated URLs
    """
    for url, snapshot in wayback_urls.items():
        # Skip cases where no Wayback Snapshot was found
        if snapshot:
            md_source = md_source.replace(url, snapshot)
    return md_source
